//HookServie.c
#include "ntddk.h"
#include "Prism.h"
#include "Intel80x86.h"
#include "UndocNTAPI.h"

typedef NTSTATUS (*NTCREATEFILE)(
	OUT PHANDLE FileHandle,
	IN ACCESS_MASK DesiredAccess,
	IN POBJECT_ATTRIBUTES ObjectAttributes,
	OUT PIO_STATUS_BLOCK IoStatusBlock,
	IN PLARGE_INTEGER AllocationSize OPTIONAL,
	IN ULONG FileAttributes,
	IN ULONG ShareAccess,
	IN ULONG CreateDisposition,
	IN ULONG CreateOption,
	IN PVOID EaBuffer OPTIONAL,
	IN ULONG EaLength
	 );

#define 		POLLING_INTERVAL 100

BOOLEAN 		Hook_Service(int nServiceID, PVOID funcHook, PVOID *oldFunc);
NTCREATEFILE 	OldNtCreateFile = NULL;
static ULONG	G_nLockUseCounter = 0;
static ULONG	SERVICEID_CREATEFILE;

NTSTATUS HookCreateFile(
	OUT PHANDLE FileHandle,
	IN ACCESS_MASK DesiredAccess,
	IN POBJECT_ATTRIBUTES ObjectAttributes,
	OUT PIO_STATUS_BLOCK IoStatusBlock,
	IN PLARGE_INTEGER AllocationSize OPTIONAL,
	IN ULONG FileAttributes,
	IN ULONG ShareAccess,
	IN ULONG CreateDisposition,
	IN ULONG CreateOption,
	IN PVOID EaBuffer OPTIONAL,
	IN ULONG EaLength
	 )
{
	NTSTATUS ntReturn;
	
	InterlockedIncrement( &G_nLockUseCounter );
	
	DbgPrint("Called CreateFile = %ws \n", 
				ObjectAttributes->ObjectName->Buffer);	
	//1. Call the original function
	ntReturn = OldNtCreateFile ( FileHandle,	DesiredAccess, 
				ObjectAttributes, IoStatusBlock , AllocationSize,
			 	FileAttributes, ShareAccess, CreateDisposition, 
			 	CreateOption, EaBuffer, EaLength );
	
	InterlockedDecrement( &G_nLockUseCounter );	
	return ntReturn;
	
}

VOID My_DriverEntry(IN PDRIVER_OBJECT DriverObject ,PDEVICE_OBJECT deviceObject)
{
	BEGINE_WIN_VER(WIN_VER_2K)
		SERVICEID_CREATEFILE = 0x20;	//Win 2k
	ELSE_WIN_VER()
		SERVICEID_CREATEFILE = 0x25;	//Win XP
	END_WIN_VER()	
}


VOID My_DriverUnload(IN PDRIVER_OBJECT DriverObject)
{
	KTIMER timer;
	LARGE_INTEGER duetime = {0};
	
	KeInitializeTimerEx( &timer, SynchronizationTimer );
	KeSetTimerEx( &timer, duetime, POLLING_INTERVAL, NULL );		
	while (G_nLockUseCounter > 0)
	{
		KeWaitForSingleObject(&timer, Executive, KernelMode, FALSE, NULL );
	}
	KeCancelTimer( &timer );	
}


NTSTATUS My_IoControl(PFILE_OBJECT pFileObject, ULONG nIoCode, PCHAR pSystemBuffer,
		 ULONG nInput, ULONG nOutput, ULONG *nReturnOut)
{
	if(nIoCode == IOCTL_HOOK_CREATE_FILE)
	{
		if(OldNtCreateFile == NULL)
			Hook_Service(SERVICEID_CREATEFILE, HookCreateFile, &(PVOID)OldNtCreateFile);		
	}	
	
	if(nIoCode == IOCTL_UNHOOK_CREATE_FILE)
	{
		if(OldNtCreateFile)
		{
			Hook_Service(SERVICEID_CREATEFILE, OldNtCreateFile, NULL);
			OldNtCreateFile = NULL;
		}		
	}	
		
	return STATUS_SUCCESS;
}

BOOLEAN Hook_Service(int nServiceID, PVOID funcHook, PVOID *oldFunc)
{
	if(oldFunc)
	{
		*oldFunc = (PVOID)KeServiceDescriptorTable.ServiceTableBase[ nServiceID ];
		
		DbgPrint("OldNtCreateFile = 0x%x, Params = 0x%x \n", *oldFunc, 
			KeServiceDescriptorTable.ParamTableBase[ nServiceID ]);
	}	
	
	KeServiceDescriptorTable.ServiceTableBase[ nServiceID ] = (unsigned int)funcHook;	
	
	return TRUE;
}
